from collections import defaultdict
from typing import Set, List, Protocol, runtime_checkable, Dict, Tuple, Iterable, TextIO

from ast_grep_py import SgNode, SgRoot
from pydantic import BaseModel, Field


class EnvVar(BaseModel):
    node: SgNode
    name: str
    position: Tuple[int, int]
    file: str = ""
    cast: str = ""
    value: str = ""
    required: bool = False

    class Config:
        arbitrary_types_allowed = True

    def location(self) -> str:
        return f"{self.file}:{self.position[0]},{self.position[1]}"

    def statement(self) -> str:
        return self.node.text()

    def merge_from(self, other: "EnvVar", strict: bool = True) -> bool:
        if self.name != other.name:
            return False

        if strict and self.file != other.file:
            return False

        if strict and self.position != other.position:
            return False

        merged = False
        if not self.cast and other.cast:
            self.cast = other.cast
            merged = True

        if not self.value and other.value:
            self.value = other.value
            merged = True

        my_range = self.node.range()
        other_range = other.node.range()

        if my_range.start.line < other_range.start.line:
            self.node = other.node
            merged = True
        elif (
            my_range.start.line == other_range.start.line
            and my_range.start.column < other_range.start.column
        ):
            self.node = other.node
            merged = True

        if my_range.end.line > other_range.end.line:
            self.node = other.node
            merged = True
        elif (
            my_range.end.line == other_range.end.line
            and my_range.end.column > other_range.end.column
        ):
            self.node = other.node
            merged = True

        return merged


class PYVar(BaseModel):
    node: SgNode
    name: str

    class Config:
        arbitrary_types_allowed = True


class Context(BaseModel):
    filename: str = ""
    imports: Set[str] = Field(default_factory=set)
    star_imports: Set[str] = Field(default_factory=set)
    env_vars: Dict[Tuple[int, int], EnvVar] = Field(default_factory=dict)

    def has_module(self, path: str) -> bool:
        module, sep, name = path.rpartition(":")

        if sep:
            return self.has_imports(module, name)

        return self.has_imports(name, "")

    def has_imports(self, path: str, name: str = "") -> bool:
        if not path:
            return True

        if path in self.imports:
            return True

        if name and f"{path}:{name}" in self.imports:
            return True

        return self.has_star_imports()

    def has_star_imports(self, path: str = "") -> bool:
        if not path:
            return bool(self.star_imports)

        return path in self.star_imports

    def add_module(self, path: str):
        self.imports.add(path)

    def add_imports(self, module: str, name: str):
        self.imports.add(f"{module}:{name}")

        if name == "*":
            self.star_imports.add(module)

    def add_env_var(self, env_var: EnvVar) -> bool:
        env_var.file = self.filename
        declared = self.env_vars.get(env_var.position)
        if declared:
            declared.merge_from(env_var)
            return False

        self.env_vars[env_var.position] = env_var
        return True

    def iter_env_vars(self) -> Iterable[EnvVar]:
        return self.env_vars.values()


class Plugin(BaseModel):
    priority: int

    def should_run(self, context: Context, node: SgNode) -> bool:
        return True

    def handle(self, context: Context, node: SgNode):
        raise NotImplementedError

    def run(self, context: Context, node: SgNode):
        if not self.should_run(context, node):
            return

        self.handle(context, node)


@runtime_checkable
class EnvVarExporter(Protocol):
    def begin(self): ...
    def handle(self, var: EnvVar): ...
    def end(self): ...


class PluginHub(BaseModel):
    exporter: EnvVarExporter

    class Config:
        arbitrary_types_allowed = True

    @property
    def plugins(self) -> List[Plugin]:
        raise NotImplementedError

    def run(self, files: Iterable[TextIO]):
        results: Dict[str, List[EnvVar]] = defaultdict(list)
        sorted_plugins = sorted(self.plugins, key=lambda p: p.priority)
        self.exporter.begin()
        for f in files:
            code = f.read()
            root = SgRoot(code, "python")

            context = Context(filename=f.name)
            root_node = root.root()
            for plugin in sorted_plugins:
                plugin.run(context, root_node)

            for v in context.iter_env_vars():
                results[v.name].append(v)

        for v in results:
            for var in results[v]:
                self.exporter.handle(var)

        self.exporter.end()



from typing import List

import click

from upcast.exporter import CSVExporter, ConsoleExporter
from upcast.plugins.env_var import EnvVarHub


@click.group()
def main():
    pass


@main.command()
@click.option("-o", "--output", default="", type=click.Path())
@click.argument("path", nargs=-1)
def find_env_vars(output: str, path: List[str]):
    def iter_files():
        for i in path:
            with open(i, "r") as f:
                yield f

    if not output:
        exporter = ConsoleExporter()
    elif output.endswith(".csv"):
        exporter = CSVExporter(path=output)
    else:
        raise click.UsageError("Output format not supported")

    hub = EnvVarHub(exporter=exporter)
    hub.run(iter_files())


if __name__ == "__main__":
    main()



import ast
from typing import Optional, List, Set

from ast_grep_py import SgNode, Range

from upcast.core import Plugin, Context, EnvVar, PluginHub
from upcast.plugins.base import ModuleImportPlugin


class FixMixin:
    value_kind_to_cast_mappings = {
        "string": "str",
        "integer": "int",
        "float": "float",
        "true": "bool",
        "false": "bool",
        "list": "list",
        "tuple": "tuple",
        "dictionary": "dictionary",
    }

    def handle_name(self, node: Optional[SgNode]) -> (str, Range):
        if not node:
            return ""

        if not node.matches(kind="string"):
            return ""

        statement = node.text()
        if statement.startswith("f"):
            # fstring is not supported
            return ""

        return ast.literal_eval(statement)

    def handle_value(
        self, cast_node: Optional[SgNode], value_node: Optional[SgNode]
    ) -> (str, str):
        cast = ""
        if cast_node and cast_node.matches(kind="identifier"):
            cast = cast_node.text()

        if not value_node:
            return cast, ""

        if not cast and value_node.kind() in self.value_kind_to_cast_mappings:
            cast = self.value_kind_to_cast_mappings[value_node.kind()]

        return cast, value_node.text()

    def make_env_var(
        self,
        result: SgNode,
        required: bool,
    ) -> Optional[EnvVar]:
        name_node = result.get_match("NAME")
        if not name_node:
            return None

        name = self.handle_name(name_node)
        if not name:
            return None

        cast, value = self.handle_value(
            result.get_match("TYPE"), result.get_match("VALUE")
        )

        name_node_range = name_node.range()

        return EnvVar(
            node=result,
            name=name,
            value=value,
            cast=cast,
            required=required,
            position=(name_node_range.start.line, name_node_range.start.column),
        )


class EnvRefPlugin(Plugin, FixMixin):
    pattern: str
    module: str = ""
    imports: str = ""
    type_convert: bool = True
    or_default: bool = True
    required: bool = False
    priority: int = 8

    @property
    def patterns(self) -> List[str]:
        yield self.pattern

        if self.type_convert and self.or_default:
            yield f"$TYPE({self.pattern}) or $VALUE"
            yield f"$TYPE({self.pattern} or $VALUE)"

        if self.type_convert:
            yield f"$TYPE({self.pattern})"

        if self.or_default:
            yield f"{self.pattern} or $VALUE"

    def should_run(self, context: Context, node: SgNode) -> bool:
        return context.has_imports(self.module, self.imports)

    def iter_var_by_pattern(self, pattern: str, node: SgNode):
        for i in node.find_all(pattern=pattern):
            env_var = self.make_env_var(i, self.required)
            if env_var:
                yield env_var

    def handle(self, context: Context, node: SgNode):
        for pattern in self.patterns:
            for i in self.iter_var_by_pattern(pattern, node):
                context.add_env_var(i)


class DjangoEnvPlugin(EnvRefPlugin, FixMixin):
    pattern: str = ""
    priority: int = 7
    var_class: str = "Env"
    var_name: str = "env"
    defined_vars: Set[str] = set()

    @property
    def patterns(self) -> List[str]:
        yield f"{self.var_name}.$TYPE($NAME)"
        yield f"{self.var_name}($NAME)"
        yield f"{self.var_name}.$TYPE($NAME, default=$VALUE)"
        yield f"{self.var_name}.$TYPE($NAME, $VALUE)"

    def should_run(self, context: Context, node: SgNode) -> bool:
        if context.has_module("environ"):
            self.var_class = "environ.Env"

            return True

        if context.has_imports("environ", "Env"):
            return True

        return False

    def handle_declare(self, context: Context, node: SgNode) -> str:
        declare_node = node.find(pattern=f"$NAME = {self.var_class}($$$ARGS)")
        if not declare_node:
            return self.var_name

        self.var_name = declare_node["NAME"].text()

        for i in declare_node.get_multiple_matches("ARGS"):
            if not i.matches(kind="keyword_argument"):
                continue

            name_node = i.child(0)
            name = name_node.text()
            name_node_range = name_node.range()
            arg_node = i.child(2)
            cast_node = arg_node.child(1)
            value_node = arg_node.child(3)

            self.defined_vars.add(name)
            context.add_env_var(
                EnvVar(
                    node=i,
                    name=name,
                    value=value_node.text(),
                    cast=cast_node.text(),
                    position=(
                        name_node_range.start.line,
                        name_node_range.start.column,
                    ),
                )
            )

    def handle(self, context: Context, node: SgNode):
        self.handle_declare(context, node)
        for pattern in self.patterns:
            for i in self.iter_var_by_pattern(pattern, node):
                if not i.value and i.name not in self.defined_vars:
                    i.required = True

                context.add_env_var(i)


class EnvVarHub(PluginHub):
    django_env_name: str = "env"

    @property
    def plugins(self) -> List[Plugin]:
        return [
            ModuleImportPlugin(),
            # stdlib
            EnvRefPlugin(pattern="os.getenv($NAME)", module="os"),
            EnvRefPlugin(pattern="os.getenv($NAME, $VALUE)", module="os"),
            EnvRefPlugin(pattern="os.environ[$NAME]", module="os", required=True),
            EnvRefPlugin(pattern="os.environ.get($NAME)", module="os"),
            EnvRefPlugin(pattern="os.environ.get($NAME, $VALUE)", module="os"),
            EnvRefPlugin(pattern="getenv($NAME)", module="os", imports="getenv"),
            EnvRefPlugin(
                pattern="getenv($NAME, $VALUE)", module="os", imports="getenv"
            ),
            EnvRefPlugin(
                pattern="environ[$NAME]", module="os", imports="environ", required=True
            ),
            EnvRefPlugin(pattern="environ.get($NAME)", module="os", imports="environ"),
            EnvRefPlugin(
                pattern="environ.get($NAME, $VALUE)", module="os", imports="environ"
            ),
            # django env
            DjangoEnvPlugin(),
        ]



import csv
from dataclasses import dataclass, field
from typing import TextIO, List, Dict

from pydantic import BaseModel

from upcast.core import EnvVar


class BaseExporter(BaseModel):
    def begin(self):
        pass

    def handle(self, env_var: EnvVar):
        raise NotImplementedError()

    def end(self):
        pass


@dataclass
class CSVExporter:
    path: str
    file: TextIO = field(init=False)
    writer: csv.DictWriter = field(init=False)

    def __post_init__(self):
        self.file = open(self.path, "w", encoding="utf-8-sig")
        self.writer = csv.DictWriter(
            self.file,
            fieldnames=["name", "cast", "value", "required", "location", "statement"],
        )

    def begin(self):
        self.writer.writeheader()

    def handle(self, env_var: EnvVar):
        self.writer.writerow(
            {
                "name": env_var.name,
                "cast": env_var.cast,
                "value": env_var.value,
                "required": "*" if env_var.required else "",
                "location": env_var.location(),
                "statement": env_var.statement(),
            }
        )

    def end(self):
        self.file.close()


class ConsoleExporter(BaseExporter):
    def handle(self, env_var: EnvVar):
        prefix = ""
        if env_var.required:
            prefix = "*"

        if env_var.value:
            print(f"{prefix}{env_var.name}={env_var.value} at {env_var.location()}")
        else:
            print(f"{prefix}{env_var.name} at {env_var.location()}")


class CollectionExporter(BaseExporter):
    collected_vars: List[EnvVar] = field(default_factory=list)

    def handle(self, env_var: EnvVar):
        self.collected_vars.append(env_var)

    def get_merged_vars(self) -> Dict[str, EnvVar]:
        merged_vars: Dict[str, EnvVar] = {}
        for i in self.collected_vars:
            if i.name in merged_vars:
                merged_vars[i.name].merge_from(i, False)
            else:
                merged_vars[i.name] = i

        return merged_vars



from ast_grep_py import SgNode

from upcast.core import Plugin, Context


class ModuleImportPlugin(Plugin):
    priority: int = 2

    def handle_import(self, context: Context, node: SgNode) -> bool:
        result = node.find(pattern=f"import $MODULE")
        if not result:
            return False

        module_node = result.get_match("MODULE")
        context.add_module(module_node.text())

    def handle_import_from(self, context: Context, node: SgNode):
        result = node.find_all(pattern="from $MODULE import $$$NAME")
        for i in result:
            module_node = i.get_match("MODULE")
            module_name = module_node.text()

            for name_node in i.get_multiple_matches("NAME"):
                if name_node.kind() == ",":
                    continue

                context.add_imports(module_name, name_node.text())

    def handle(self, context: Context, node: SgNode):
        self.handle_import(context, node)
        self.handle_import_from(context, node)


class PyVarPlugin(Plugin):
    priority: int = 2

    def handle(self, context: Context, node: SgNode):
        if node.matches(kind="assign"):
            name = node.get_match("NAME").text()
            context.add_py_var(name)



from dataclasses import field
from typing import List, Dict, Optional

from ast_grep_py import SgNode


class FunctionArgs:
    args: List[SgNode] = field(default_factory=list)
    kwargs: Dict[str, SgNode] = field(default_factory=dict)

    def parse(self, node: SgNode, group: str):
        for i in node.get_multiple_matches(group):
            if i.matches(kind=","):
                continue
            elif i.matches(kind="keyword_argument"):
                self.kwargs[i.child(0).text()] = i.child(1)
            else:
                self.args.append(i)

    def get_args(self, keys: Dict[str, int]) -> Dict[str, Optional[SgNode]]:
        args: Dict[str, Optional[SgNode]] = {}
        for key, index in keys.items():
            if index < len(self.args):
                args[key] = self.args[index]
            elif key in self.kwargs:
                args[key] = self.kwargs[key]

        return args
